#include <cmath>
#include <boost/bind.hpp>
#include <gazebo_msgs/DeleteModel.h>
#include <gazebo/gazebo.hh>
#include <gazebo/physics/physics.hh>
#include <gazebo/common/common.hh>
#include <gazebo/transport/transport.hh>
#include <ros/ros.h>
#include <actionlib/server/simple_action_server.h>
#include <tf/transform_listener.h>
#include <iri_common_drivers_msgs/tool_closeAction.h>
#include <iri_common_drivers_msgs/tool_openAction.h>

namespace gazebo
{

class BlackHolePlugin : public ModelPlugin
{
  private:
    transport::SubscriberPtr contactSub_;
    physics::ModelPtr model_;
    physics::LinkPtr link_;
    physics::ContactManager* contact_manager_;
    math::Pose trash_;
    transport::NodePtr node_;
    std::vector<std::string> collision_names_, stars_;
    std::string topic_; // contacts topic

    void read_options(sdf::ElementPtr sdf)
    {
      sdf::ElementPtr elem;

      // Read black hole link
      if (sdf->HasElement("link"))
      {
        elem = sdf->GetElement("link");
        std::string link_name = elem->GetValue()->GetAsString();
        link_ = model_->GetLink(link_name);
      }
      
      if (sdf->HasElement("star"))
      {
        elem = sdf->GetElement("star");
      }

      while (elem)
      {
        std::string star = elem->GetValue()->GetAsString();
        stars_.push_back(star);
        elem = elem->GetNextElement("star");
        ROS_INFO("Added star: %s", star.c_str());
      }

      bool ok = false;

      elem.reset();

      if (sdf->HasElement("trash"))
      {
        elem = sdf->GetElement("trash");
      }

      if (elem)
      {
        sdf::Pose trash;
        ok = elem->GetValue()->Get(trash);
        if (ok)
        {
          trash_.pos.x = trash.pos.x;
          trash_.pos.y = trash.pos.y;
          trash_.pos.z = trash.pos.z;
          trash_.rot.w = trash.rot.w;
          trash_.rot.x = trash.rot.x;
          trash_.rot.y = trash.rot.y;
          trash_.rot.z = trash.rot.z;
        }
        else
        {
          ROS_WARN("Invalid content inside the trash's tag.");
        }
      }

      if (not elem or not ok)
      {
        ROS_WARN("Using default trash pose: -3 -2 1.5 0 0 0");
        math::Vector3 pos(-3, -2, 1.5);
        math::Quaternion rot(0, 0, 0, 1);
        trash_ = math::Pose(pos, rot);
      }
    }
    
    /** 
     * @brief Iterate through all the collisions of the model to get their
     * names and store them in collision_names_
     */
    void obtain_collision_names()
    {
      //collision_names_.clear();
      if (link_)
      {
        auto collisions = link_->GetCollisions();
        for (auto collision : collisions)
        {
          collision_names_.push_back(collision->GetScopedName());
        }
      }
      else
      {
        physics::Link_V links = model_->GetLinks();
        for (auto link : links)
        {
          physics::Collision_V collisions = link->GetCollisions();
          for (auto collision : collisions)
          {
            collision_names_.push_back(collision->GetScopedName());
          }
        }
      }
    }

    bool is_star(physics::ModelPtr model)
    {
      for (int i = 0; i < stars_.size(); ++i)
      {
        if (stars_[i] == model->GetName()) return true;
      }
      return false;
    }

  public:

    virtual void Load(physics::ModelPtr parent, sdf::ElementPtr sdf)
    {
      model_ = parent;

      // get contact manager from physics engine
      physics::PhysicsEnginePtr physics =
        parent->GetWorld()->GetPhysicsEngine();
      contact_manager_ = physics->GetContactManager();
      read_options(sdf);
      obtain_collision_names();
      topic_ = contact_manager_->CreateFilter(model_->GetName() + "_contacts",
          collision_names_);

      // init gazebo node and link contact callback
      node_.reset(new transport::Node());
      node_->Init(model_->GetWorld()->GetName());
      contactSub_ = node_->Subscribe(topic_, &BlackHolePlugin::OnContact, this);
    }

    void OnContact(const ConstContactsPtr& msg)
    {
      //ROS_INFO("New contact");
      for (int i = 0; i < msg->contact_size(); ++i)
      {
        physics::CollisionPtr collision1 = boost::dynamic_pointer_cast<
          physics::Collision>(model_->GetWorld()->GetEntity(
                msg->contact(i).collision1()));
        physics::CollisionPtr collision2 = boost::dynamic_pointer_cast<
          physics::Collision>(model_->GetWorld()->GetEntity(
                msg->contact(i).collision2()));

        // check that the collisions exist (their models could have been
        // deleted in previous executions of this callback)
        if (not collision1 or not collision2)
        {
          continue;
        }

        physics::ModelPtr model1 = collision1->GetModel();
        physics::ModelPtr model2 = collision2->GetModel();
        physics::ModelPtr to_delete;

        // Check if any of the two models involved in the collision is a star.
        // If so, delete the star and exit the loop (maximum one deletion per
        // callback to avoid deleting the same model twice)
        //std::string to_delete;
        if (is_star(model1)) to_delete = model1;
        else if (is_star(model2)) to_delete = model2;

        if (to_delete)
        {
          ROS_INFO("Moving away model %s", to_delete->GetName().c_str());
          to_delete->SetWorldPose(trash_);
          //transport::requestNoReply(node_, "entity_delete", to_delete);
          break; // do not delete more than one model per callback invokation
        }

      }
    }

    ~BlackHolePlugin()
    {
    }

};

GZ_REGISTER_MODEL_PLUGIN(BlackHolePlugin)

}
