#include <cmath>
#include <boost/bind.hpp>
#include <gazebo/gazebo.hh>
#include <gazebo/physics/physics.hh>
#include <gazebo/common/common.hh>
#include <ros/ros.h>
#include <actionlib/server/simple_action_server.h>
#include <tf/transform_listener.h>
#include <iri_common_drivers_msgs/tool_closeAction.h>
#include <iri_common_drivers_msgs/tool_openAction.h>
#include <gz_gripper_plugin/CheckGrasped.h>

namespace gazebo
{

class GripperPlugin : public ModelPlugin
{
  private:
    ros::NodeHandle* nh_;
    event::ConnectionPtr updateConnection_;
    physics::ModelPtr model_;
    physics::LinkPtr gripper_;
    physics::LinkPtr attached_graspable_;
    math::Pose graspable_rel_pose_;
    bool deactivate_collision_;

    //physics::JointPtr fixed_joint_;
    bool closed_;

    // Initially the graspable candidate was chosen by means of a RayShape and
    // the GetIntersection method
    // physics::RayShapePtr ray_;

    // If a TF transform is not provided, use this:
    double z_offset_;

    // If a TF transform is provided, use this:
    std::string gripper_frame_id_;
    tf::TransformListener* listener_;

    // action servers
    actionlib::SimpleActionServer<iri_common_drivers_msgs::tool_closeAction>*
      close_as_;
    actionlib::SimpleActionServer<iri_common_drivers_msgs::tool_openAction>*
      open_as_;

    // service servers
    ros::ServiceServer check_ss_;

    template <typename T>
    static void set_option(sdf::ElementPtr sdf, const std::string& option,
        T& value, const T& default_)
    {
      // This template does not work for T=std::string
      if (sdf->HasElement(option))
      {
        sdf::ElementPtr elem = sdf->GetElement(option);
        if (elem->GetValue()->Get(value))
        {
          ROS_INFO_STREAM(option << " set to " << value);
        }
        else
        {
          ROS_ERROR_STREAM(option << ": bad type. Set to default (" <<
              default_ << ")");
          value = default_;
        }
      }
      else
      {
        value = default_;
        ROS_WARN_STREAM(option << " set to default (" << default_ << ")");
      }
    }

    static bool is_contained(const math::Vector3& point, const math::Box& box)
    {
      math::Vector3 center = box.GetCenter();
      math::Vector3 size_2 = box.GetSize()/2;
      double delta_x = std::fabs(center.x - point.x);
      double delta_y = std::fabs(center.y - point.y);
      double delta_z = std::fabs(center.z - point.z);
      return delta_x < size_2.x and delta_y < size_2.y and delta_z < size_2.z;
    }

    void read_options(sdf::ElementPtr sdf)
    {
      // Read gripper link
      sdf::ElementPtr elem;
      
      elem = sdf->GetElement("gripper_link");
      if (elem)
      {
        std::string gripper_name = elem->GetValue()->GetAsString();
        ROS_INFO("Gripper link: %s", gripper_name.c_str());
        gripper_ = model_->GetLink(gripper_name);
      }
      else
      {
        ROS_ERROR("No option specifying the gripper link");
        return;
      }

      if (sdf->HasElement("gripper_tf"))
      {
        elem = sdf->GetElement("gripper_tf");
        listener_ = new tf::TransformListener;
        gripper_frame_id_ = elem->GetValue()->GetAsString();
        ROS_INFO("The position of the gripper will be inferred from the %s tf",
            gripper_frame_id_.c_str());
      }
      else
      {
        set_option(sdf, "z_offset", z_offset_, 0.5);
        //set_option(sdf, "z_reach", z_reach_, 1.0);
        listener_ = NULL;
        ROS_INFO("The position of the gripper will be inferred from the "
            "pose of the gripper link (%s) inside Gazebo",
            gripper_->GetScopedName().c_str());
      }

      set_option(sdf, "deactivate_collision", deactivate_collision_, true);

    }

    math::Vector3 get_gripper_position() const
    {
      math::Vector3 position;

      if (listener_ != NULL)
      {
        ROS_INFO("Trying to obtain gripper position through TF...");
        try
        {
          tf::StampedTransform transform;
          listener_->lookupTransform("/world", gripper_frame_id_, ros::Time(0),
              transform);
          position.x = transform.getOrigin().getX();
          position.y = transform.getOrigin().getY();
          position.z = transform.getOrigin().getZ();
        }
        catch (tf::TransformException& ex)
        {
          ROS_ERROR("%s", ex.what());
        } 
      }
      else
      {
        ROS_INFO("Obtaining gripper position from link...");
        math::Pose gripper_pose = math::Pose(math::Vector3(0,0,z_offset_),
            math::Quaternion()) + gripper_->GetWorldPose();
        position = gripper_pose.pos;
      }

      ROS_INFO_STREAM("Retrieved position: " << position);

      return position;
    }

    physics::LinkPtr get_graspable() const
    {
      physics::LinkPtr graspable;
      math::Vector3 gripper_pos = get_gripper_position();

      physics::Model_V all_models = model_->GetWorld()->GetModels();
      for (int i = 0; i < all_models.size(); ++i)
      {
        if (all_models[i] == model_) continue; // Skip this model's links

        physics::Link_V links = all_models[i]->GetLinks();
        for (int j = 0; j < links.size(); ++j)
        {
          if (is_contained(gripper_pos, links[j]->GetBoundingBox()))
          {
            graspable = links[j];
            break;
          }
        }
      }

      if (graspable)
      {
        ROS_INFO("Found graspable: %s", graspable->GetScopedName().c_str());
      }
      else
      {
        ROS_WARN("No graspable found");
      }

      return graspable;

      // This alternate version uses RayShape. It uses the GetIntersection
      // method. Sometimes the GetIntersection method crashed (that's why I'm
      // using the previous iteration).
      
      //math::Pose start = math::Pose(math::Vector3(0,0,z_offset_-0.005),
          //math::Quaternion());
      //start += gripper_->GetWorldPose();

      //math::Pose end = math::Pose(math::Vector3(0,0,0.2),
          //math::Quaternion());
      //end += start;

      //ROS_INFO_STREAM("Ray start point: " << start.pos <<
          //"; ray end point: " << end.pos);

      //ray_->SetPoints(start.pos, end.pos);

      //double dist;
      //std::string entity_n;
      //ray_->GetIntersection(dist, entity_n);

      //if (entity_n.empty()) ROS_WARN("No intersection");
      //else
      //{
        //ROS_INFO("Intersection at %fm with entity %s", dist, entity_n.c_str());
        //physics::CollisionPtr collision =
          //boost::dynamic_pointer_cast<physics::Collision>(
            //model_->GetWorld()->GetEntity(entity_n));

        //graspable = collision->GetLink();
      //}
      //return graspable;
    }

    void open_callback(
        const iri_common_drivers_msgs::tool_openGoalConstPtr& /* goal */)
    {
      if (closed_)
      {
        //physics::LinkPtr parent, child;
        //parent = fixed_joint_->GetParent();
        //child = fixed_joint_->GetChild();
        if (attached_graspable_)
        {
          ROS_INFO("Links %s and %s are connected. Detaching...",
              gripper_->GetScopedName().c_str(),
              attached_graspable_->GetScopedName().c_str());

          if (deactivate_collision_)
          {
            attached_graspable_->SetCollideMode("all");
          }

          attached_graspable_.reset();
        }
        else
        {
          ROS_INFO("No links to detach");
        }
        closed_ = false;
      }
      else
      {
        ROS_WARN("Gripper was already open. Not doing anything.");
      }

      // No errors expected => always succeed
      iri_common_drivers_msgs::tool_openResult result;
      result.successful = true;
      open_as_->setSucceeded(result);
    }

    void close_callback(
        const iri_common_drivers_msgs::tool_closeGoalConstPtr& /* goal */)
    {
      if (closed_)
      {
        ROS_WARN("Gripper was already closed. Not doing anything.");
      }
      else
      {
        attached_graspable_ = get_graspable();

        if (attached_graspable_)
        {
          ROS_INFO("Attaching %s to %s", gripper_->GetScopedName().c_str(),
              attached_graspable_->GetScopedName().c_str());
          graspable_rel_pose_ = attached_graspable_->GetWorldPose() -
            gripper_->GetWorldPose();

          if (deactivate_collision_)
          {
            attached_graspable_->SetCollideMode("none");
          }
        }
        else
        {
          ROS_WARN("The gripper has not picked anything");
        }

        closed_ = true;
      }

      // No errors expected => always succeed
      iri_common_drivers_msgs::tool_closeResult result;
      result.successful = true;
      close_as_->setSucceeded(result);
    }

    bool check_grasped(gz_gripper_plugin::CheckGrasped::Request&,
                       gz_gripper_plugin::CheckGrasped::Response& res)
    {
      res.grasped = attached_graspable_ ? attached_graspable_->GetScopedName() : ""; 
      return true;
    }

  public:
    virtual void Load(physics::ModelPtr parent, sdf::ElementPtr sdf)
    {
      model_ = parent;
      read_options(sdf);
      if (not gripper_)
      {
        // The plugin won't work. Skip the remaining initialization steps.
        ROS_ERROR("Gripper not found. The plugin will not work.");
        return;
      }

      //physics::PhysicsEnginePtr physics = parent->GetWorld()->
        //GetPhysicsEngine();
      //ray_ = boost::dynamic_pointer_cast<physics::RayShape>(
          //physics->CreateShape("ray", physics::CollisionPtr()));
      //fixed_joint_ = physics->CreateJoint("revolute", model_);

      if (not ros::isInitialized())
      {
        nh_ = NULL;
        open_as_ = NULL;
        close_as_ = NULL;
        ROS_ERROR("ROS has not been initialized. Cannot load plugin.");
        return;
      }

      nh_ = new ros::NodeHandle;
      open_as_ = new actionlib::SimpleActionServer<
        iri_common_drivers_msgs::tool_openAction>(*nh_, "gripper/open_tool",
            boost::bind(&GripperPlugin::open_callback, this, _1), false);
      close_as_ = new actionlib::SimpleActionServer<
        iri_common_drivers_msgs::tool_closeAction>(*nh_, "gripper/close_tool",
            boost::bind(&GripperPlugin::close_callback, this, _1), false);

      check_ss_ = nh_->advertiseService("check_grasped",
                                        &GripperPlugin::check_grasped,
                                        this);

      open_as_->start();
      close_as_->start();      

      ROS_INFO("Gripper plugin loaded");

      this->updateConnection_ = event::Events::ConnectWorldUpdateBegin(
          boost::bind(&GripperPlugin::OnUpdate, this, _1));
    }

    void OnUpdate(const common::UpdateInfo info)
    {
      if (attached_graspable_)
      {
        attached_graspable_->SetWorldPose(graspable_rel_pose_ +
            gripper_->GetWorldPose());
        attached_graspable_->SetWorldTwist(math::Vector3(), math::Vector3());
      }
      ros::spinOnce();
    }

    ~GripperPlugin()
    {
      if (nh_ != NULL) delete nh_;
      if (open_as_ != NULL) delete open_as_;
      if (close_as_ != NULL) delete close_as_;
      if (listener_ != NULL) delete listener_;
    }

};

GZ_REGISTER_MODEL_PLUGIN(GripperPlugin)

}
